/* ARCADEmini USB driver for Windows
 * version 1.0.0 date 2004-08-25
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "AmUsbInternal.h"

//Used as a  general purpose completion routine so it can signal an event,
//passed as the Context, when the next lower driver is done with the input Irp.
//This routine is used by both PnP and Power Management logic.
//pDeviceObject: pointer to the device object for the class device
//pIrp: the IRP completed
//pContext: driver defined context, in this case a pointer to an event
//return value: final status from the operation
NTSTATUS AmUsbIrpCompletionRoutine( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp, IN PVOID pContext )
{
	PKEVENT evEvent = (PKEVENT)pContext;

	//set the input event
	KeSetEvent( evEvent, 1, //priority increment for waiting thread
			FALSE );  //this call is not immediately followed by wait

	//This routine must return STATUS_MORE_PROCESSING_REQUIRED because
	//we have not yet called IoFreeIrp() on this IRP.
	return STATUS_MORE_PROCESSING_REQUIRED;
}

//Called from AmUsbPnp(), the dispatch routine for IRP_MJ_PNP.
//Initializes a given instance of the device on the USB.
//USB client drivers such as us set up URBs (USB Request Packets) to send requests
//to the host controller driver (HCD). The URB structure defines a format for all
//possible commands that can be sent to a USB device.
//Here, we request the device descriptor and store it, and configure the device.
//pDeviceObject: pointer to the FDO (Functional Device Object)
//return value: NT status code
NTSTATUS AmUsbStartDevice( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status;
	PUSB_DEVICE_DESCRIPTOR pDeviceDescriptor = NULL;
	PURB pUrb;
	ULONG Siz;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbStartDevice\r\n" ) );

	//get device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//allocate a URB
	pUrb = ExAllocatePool( NonPagedPool, sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST ) );

	//debug message, in case of error
	AmUsbDebugCond( DBGLVL_HIGH, !pUrb, ( "AmUsbStartDevice() FAILED, ExAllocatePool() for URB FAILED\r\n" ) );

	//we got the URB
	if( pUrb )
	{
		//allocate memory for the device descriptor
		Siz = sizeof( USB_DEVICE_DESCRIPTOR );
		pDeviceDescriptor = ExAllocatePool( NonPagedPool, Siz );

		//debug message, in case of error
		AmUsbDebugCond( DBGLVL_HIGH, !pDeviceDescriptor, ( "AmUsbStartDevice() FAILED, ExAllocatePool() for pDeviceDescriptor FAILED\r\n" ) );

		//we got the memory for the device descriptor
		if( pDeviceDescriptor )
		{
			//build the USB request for the device-descriptor
			UsbBuildGetDescriptorRequest( pUrb,
					sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST ),
					USB_DEVICE_DESCRIPTOR_TYPE,
					0, 0, pDeviceDescriptor, NULL, Siz, NULL );
			//submit the request
			Status = AmUsbCallUsbd( pDeviceObject, pUrb );
			//debug message, in case of error
			AmUsbDebugCond( DBGLVL_DEFAULT, !NT_SUCCESS( Status ), ( "AmUsbStartDevice() FAILED, AmUsbCallUSBD( pDeviceObject, pUrb ) FAILED\r\n" ) );
			//success
			if( NT_SUCCESS( Status ) )
			{
				//debug messages with device descriptor
				AmUsbDebug( DBGLVL_MEDIUM, ( "Device Descriptor = %x, len 0x%x\r\n",
						pDeviceDescriptor, pUrb->UrbControlDescriptorRequest.TransferBufferLength ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bLength %d\r\n", pDeviceDescriptor->bLength ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bDescriptorType 0x%x\r\n", pDeviceDescriptor->bDescriptorType ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bcdUSB 0x%x\r\n", pDeviceDescriptor->bcdUSB ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bDeviceClass 0x%x\r\n", pDeviceDescriptor->bDeviceClass ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bDeviceSubClass 0x%x\r\n", pDeviceDescriptor->bDeviceSubClass ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bDeviceProtocol 0x%x\r\n", pDeviceDescriptor->bDeviceProtocol ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bMaxPacketSize0 0x%x\r\n", pDeviceDescriptor->bMaxPacketSize0 ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "idVendor 0x%x\r\n", pDeviceDescriptor->idVendor ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "idProduct 0x%x\r\n", pDeviceDescriptor->idProduct ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bcdDevice 0x%x\r\n", pDeviceDescriptor->bcdDevice ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "iManufacturer 0x%x\r\n", pDeviceDescriptor->iManufacturer ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "iProduct 0x%x\r\n", pDeviceDescriptor->iProduct ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "iSerialNumber 0x%x\r\n", pDeviceDescriptor->iSerialNumber ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "bNumConfigurations 0x%x\r\n", pDeviceDescriptor->bNumConfigurations ) );
			}
		}
		//we got no memory for the device descriptor
		else
			//set error code to indicate insufficient resources
			Status = STATUS_INSUFFICIENT_RESOURCES;

		//success
		if( NT_SUCCESS( Status ) )
			//save the device descriptor
			pDeviceExtension->pUsbDeviceDescriptor = pDeviceDescriptor;
		//error, but device descriptor allocated
		else if( pDeviceDescriptor )
			//free memory of device descriptor
			ExFreePool( pDeviceDescriptor );

		//free memory of URB
		ExFreePool( pUrb );
	}
	//we got no memory for the URB
	else
		//set error code to indicate insufficient resources
		Status = STATUS_INSUFFICIENT_RESOURCES;

	//success
	if( NT_SUCCESS( Status ) )
	{
		//configure the device
		Status = AmUsbConfigureDevice( pDeviceObject );
		//debug message, in case of error
		AmUsbDebugCond( DBGLVL_MEDIUM, !NT_SUCCESS( Status ), ( "AmUsbStartDevice, AmUsbConfigureDevice() FAILED with 0x%x\r\n", Status ) );
	}

	//still success
	if( NT_SUCCESS( Status ) )
		//remember that device is started
		pDeviceExtension->DeviceStarted = TRUE;

	//no private device data yet
	pDeviceExtension->pDev = NULL;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ("exiting AmUsbStartDevice with status 0x%x\r\n", Status ) );

	return Status;
}

//Stops a given instance of a AmUsb device on the USB.
//We basically just tell USB this device is now 'unconfigured'
//pDeviceObject: pointer to the device object for this instance
//return value: NT status code
NTSTATUS AmUsbStopDevice( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status = STATUS_SUCCESS;
	PURB pUrb;
	USHORT Siz;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "enter AmUsbStopDevice\r\n" ) );

	//get device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//Send the select configuration URB with a NULL pointer for the configuration
	//handle. This closes the configuration and puts the device in the 'unconfigured'
	//state.
	Siz = sizeof( struct _URB_SELECT_CONFIGURATION );
	//allocate URB
	pUrb = ExAllocatePool( NonPagedPool, Siz );
	//we got the memory
	if( pUrb )
	{
		//build select configuration 0 request
		UsbBuildSelectConfigurationRequest( pUrb, Siz, NULL );
		//submit the URB
		Status = AmUsbCallUsbd( pDeviceObject, pUrb );

		//debug message, in case of error
		AmUsbDebugCond( DBGLVL_DEFAULT, !NT_SUCCESS( Status ), ( "AmUsbStopDevice() FAILED, Configuration Closed status = 0x%x, USB status = 0x%x\r\n", Status, pUrb->UrbHeader.Status ) );

		//free URB
		ExFreePool( pUrb );
	}
	//we got no memory for the URB
	else
		//set error code to indicate insufficient resources
		Status = STATUS_INSUFFICIENT_RESOURCES;

	//success
	if( NT_SUCCESS( Status ) )
		//remember that device is stopped
		pDeviceExtension->DeviceStarted = FALSE;

	//free device descriptor structure
	if( pDeviceExtension->pUsbDeviceDescriptor )
	{
		ExFreePool( pDeviceExtension->pUsbDeviceDescriptor );
		pDeviceExtension->pUsbDeviceDescriptor = NULL;
	}

	//free up the USB config discriptor
	if( pDeviceExtension->pUsbConfigurationDescriptor )
	{
		ExFreePool( pDeviceExtension->pUsbConfigurationDescriptor );
		pDeviceExtension->pUsbConfigurationDescriptor = NULL;
	}

	//stopping the device is no more requested
	pDeviceExtension->StopDeviceRequested = FALSE;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbStopDevice() with status 0x%x\r\n", Status ) );

	return Status;
}

//Called from AmUsbPnp() to
//clean up our device instance's allocated buffers; free symbolic links
//pDeviceObject: pointer to the FDO
//return value: NT status code from free symbolic link operation
NTSTATUS AmUsbRemoveDevice( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status = STATUS_SUCCESS;
	UNICODE_STRING DeviceLinkUnicodeString;
	ANSI_STRING SymLinkAnsiString;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbRemoveDevice\r\n" ) );

	//get device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//symlink is set
	if( pDeviceExtension->SymLinkSet )
	{
		//debug message
		SymLinkAnsiString.Buffer = NULL;
		RtlUnicodeStringToAnsiString( &SymLinkAnsiString, &pDeviceExtension->SymLink, TRUE );
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbRemoveDevice: removing symlink \"%s\"\r\n", SymLinkAnsiString.Buffer ) );
		RtlFreeAnsiString( &SymLinkAnsiString );
		//remove symlink
		IoDeleteSymbolicLink( &pDeviceExtension->SymLink );
		//free unicode string
		RtlFreeUnicodeString( &pDeviceExtension->SymLink );
		//symlink is not set
		pDeviceExtension->SymLinkSet = FALSE;
	}

	//remove the GUID-based symbolic link
	RtlInitUnicodeString( &DeviceLinkUnicodeString, pDeviceExtension->DeviceLinkNameBuffer );
	Status = IoSetDeviceInterfaceState( &DeviceLinkUnicodeString, FALSE );
	//check that removing the link succeeded
	AmUsbAssert( NT_SUCCESS( Status ) );

	//free device descriptor structure
	if( pDeviceExtension->pUsbDeviceDescriptor)
		ExFreePool( pDeviceExtension->pUsbDeviceDescriptor );

	//free up the UsbInterface structure
	if( pDeviceExtension->pUsbInterface )
		ExFreePool( pDeviceExtension->pUsbInterface );

	//free up the USB config discriptor
	if( pDeviceExtension->pUsbConfigurationDescriptor )
		ExFreePool( pDeviceExtension->pUsbConfigurationDescriptor );

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbRemoveDevice() with 0x%x\r\n", Status ) );

	return Status;
}

//process a plug and play IRP sent to this device
//pDeviceObject: pointer to the functional device object
//pIrp: pointer to the IRP
//return value: NT status code
NTSTATUS AmUsbPnp( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status = STATUS_SUCCESS;
	NTSTATUS WaitStatus;
	PDEVICE_OBJECT pStackDeviceObject;
	KEVENT StartDeviceEvent;

	//get pointer to current location in IRP
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	//get pointer to device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;
	//get pointer to stack device
	pStackDeviceObject = pDeviceExtension->pTopOfStackDeviceObject;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "entering AmUsbPnp(), major IRP_MJ_PNP, minor %s\r\n",
			AmUsbStringForPnpMnFunc( pIrpStack->MinorFunction ) ) );
	//check major of IRP
	AmUsbAssert( pIrpStack->MajorFunction == IRP_MJ_PNP );

	//increment the FDO device extension's pending IO count for this IRP
	AmUsbIncIoCount( pDeviceObject );

	switch( pIrpStack->MinorFunction )
	{
	case IRP_MN_START_DEVICE:
		//The PnP Manager sends this IRP after it has assigned resources, 
		//if any, to the device. The device may have been recently enumerated
		//and is being started for the first time, or the device may be 
		//restarting after being stopped for resource reconfiguration. 

		//initialize an event we can wait on for the PDO to be done with this IRP
		KeInitializeEvent( &StartDeviceEvent, NotificationEvent, FALSE );

		//copy IRP to next one
		IoCopyCurrentIrpStackLocationToNext( pIrp );

		//set completion routine so PDO driver can signal event when PDO is done with the IRP
		IoSetCompletionRoutine( pIrp, AmUsbIrpCompletionRoutine, &StartDeviceEvent, TRUE, TRUE, TRUE );

		//let PDO process the IRP
		Status = IoCallDriver( pStackDeviceObject, pIrp );

		//if PDO is not done yet, wait for event to be set in our completion routine
		if( Status == STATUS_PENDING )
			//wait for IRP to complete
			WaitStatus = KeWaitForSingleObject( &StartDeviceEvent, Suspended, KernelMode, FALSE, NULL );

		//setup the device and start it
		Status = AmUsbStartDevice( pDeviceObject );
		//save state
		pIrp->IoStatus.Status = Status;

		//complete request
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );
		//decrement the FDO device extension's pending IO count for this IRP
		AmUsbDecIoCount( pDeviceObject );

		//start real driver
		AmUsbProbe( pDeviceObject->DeviceExtension );

		return Status;

	case IRP_MN_QUERY_STOP_DEVICE:
		//The IRP_MN_QUERY_STOP_DEVICE/IRP_MN_STOP_DEVICE sequence only occurs
		//during "polite" shutdowns, such as the user explicitily requesting the
		//service be stopped in, or requesting unplug from the Pnp tray icon.
		//This sequence is NOT received during "impolite" shutdowns,
		//such as someone suddenly yanking the USB cord or otherwise 
		//unexpectedly disabling/resetting the device.

		//If a driver sets STATUS_SUCCESS for this IRP,
		//the driver must not start any operations on the device that
		//would prevent that driver from successfully completing an IRP_MN_STOP_DEVICE
		//for the device.
		//For mass storage devices such as disk drives, while the device is in the
		//stop-pending state, the driver holds IRPs that require access to the device,
		//but for most USB devices, there is no 'persistent storage', so we will just
		//refuse any more IO until restarted or the stop is cancelled

		//If a driver in the device stack determines that the device cannot be
		//stopped for resource reconfiguration, the driver is not required to pass
		//the IRP down the device stack. If a query-stop IRP fails,
		//the PnP Manager sends an IRP_MN_CANCEL_STOP_DEVICE to the device stack,
		//notifying the drivers for the device that the query has been cancelled
		//and that the device will not be stopped.

		//It is possible to receive this IRP when the device has not been started
		//(as on a boot device)
		if( !pDeviceExtension->DeviceStarted )
		{
			//debug message
			AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPnp() IRP_MN_QUERY_STOP_DEVICE when device not started\r\n" ) );
			//just pass to lower driver
			IoSkipCurrentIrpStackLocation( pIrp );
			Status = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );
			//decrement the FDO device extension's pending IO count for this IRP
			AmUsbDecIoCount( pDeviceObject );
			return Status;
		}

		//stop device is now requested
		pDeviceExtension->StopDeviceRequested = TRUE;
		break;

	case IRP_MN_CANCEL_STOP_DEVICE:
		//The PnP Manager uses this IRP to inform the drivers for a device
		//that the device will not be stopped for resource reconfiguration.
		//This should only be received after a successful IRP_MN_QUERY_STOP_DEVICE.

		//It is possible to receive this IRP when the device has not been started
		if( !pDeviceExtension->DeviceStarted )
		{
			//debug message
			AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPnp() IRP_MN_CANCEL_STOP_DEVICE when device not started\r\n" ) );
			//just pass to lower driver
			IoSkipCurrentIrpStackLocation( pIrp );
			Status = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );
			//decrement the FDO device extension's pending IO count for this IRP
			AmUsbDecIoCount( pDeviceObject );
			return Status;
		}

		//reset this flag so new IOCTL and IO IRP processing will be re-enabled
		pDeviceExtension->StopDeviceRequested = FALSE;
		break;

	case IRP_MN_STOP_DEVICE:
		//The PnP Manager sends this IRP to stop a device so it can reconfigure
		//its hardware resources. The PnP Manager only sends this IRP if a prior
		//IRP_MN_QUERY_STOP_DEVICE completed successfully.

		//stop real driver
		AmUsbDisconnect( pDeviceObject->DeviceExtension );

		//stop the device
		Status = AmUsbStopDevice( pDeviceObject );

		break;

	case IRP_MN_QUERY_REMOVE_DEVICE:
		//In response to this IRP, drivers indicate whether the device can be
		//removed without disrupting the system.
		//If a driver determines it is safe to remove the device,
		//the driver completes any outstanding I/O requests, arranges to hold any subsequent
		//read/write requests, and sets Irp->IoStatus.Status to STATUS_SUCCESS. Function
		//and filter drivers then pass the IRP to the next-lower driver in the device stack.
		//The underlying bus driver calls IoCompleteRequest.

		//If a driver sets STATUS_SUCCESS for this IRP, the driver must not start any
		//operations on the device that would prevent that driver from successfully completing
		//an IRP_MN_REMOVE_DEVICE for the device. If a driver in the device stack determines
		//that the device cannot be removed, the driver is not required to pass the
		//query-remove IRP down the device stack. If a query-remove IRP fails, the PnP Manager
		//sends an IRP_MN_CANCEL_REMOVE_DEVICE to the device stack, notifying the drivers for
		//the device that the query has been cancelled and that the device will not be removed.

		//It is possible to receive this IRP when the device has not been started
		if( !pDeviceExtension->DeviceStarted )
		{
			//debug message
			AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPnp() IRP_MN_QUERY_REMOVE_DEVICE when device not started\r\n" ) );
			//just pass to lower driver
			IoSkipCurrentIrpStackLocation( pIrp );
			Status = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );
			//decrement the FDO device extension's pending IO count for this IRP
			AmUsbDecIoCount( pDeviceObject );
			return Status;
		}

 		//Once RemoveDeviceRequested is set no new IOCTL or read/write IRPs will be passed
		//down the stack to lower drivers; all will be quickly failed
		pDeviceExtension->RemoveDeviceRequested = TRUE;

		//Wait for any IO request pending in our driver to
		//complete before returning success.
		//This event is set when pDeviceExtension->PendingIoCount goes to 1
		WaitStatus = KeWaitForSingleObject( &pDeviceExtension->evNoPendingIo,
				Suspended, KernelMode, FALSE, NULL );
		break;

	case IRP_MN_CANCEL_REMOVE_DEVICE:
		//The PnP Manager uses this IRP to inform the drivers
		//for a device that the device will not be removed.
		//It is sent only after a successful IRP_MN_QUERY_REMOVE_DEVICE.

		//It is possible to receive this IRP when the device has not been started
		if( !pDeviceExtension->DeviceStarted )
		{
			//debug message
			AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPnp() IRP_MN_CANCEL_REMOVE_DEVICE when device not started\r\n" ) );
			//just pass to lower driver
			IoSkipCurrentIrpStackLocation( pIrp );
			Status = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );
			//decrement the FDO device extension's pending IO count for this IRP
			AmUsbDecIoCount( pDeviceObject );
			return Status;
		}

		//Reset this flag so new IOCTL and IO Irp processing will be re-enabled
		pDeviceExtension->RemoveDeviceRequested = FALSE;
		break;

	case IRP_MN_SURPRISE_REMOVAL:
		//debug message
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPnp() IRP_MN_SURPRISE_REMOVAL\r\n" ) );
		
		//For a surprise-style device removal (i.e. sudden cord yank),
		//the physical device has already been removed so the PnP Manager sends
		//the remove IRP without a prior query-remove. A device can be in any state
		//when it receives a remove IRP as a result of a surprise-style removal.

		//stop real driver
		AmUsbDisconnect( pDeviceObject->DeviceExtension );

		//decrement the FDO device extension's pending IO count for this IRP
		AmUsbDecIoCount( pDeviceObject );

		//Once DeviceRemoved is set no new IOCTL or read/write IRPs will be passed
		//down the stack to lower drivers; all will be quickly failed
		pDeviceExtension->DeviceRemoved = TRUE;

/* removed because this only generated bluescreens
		//We don't explicitly wait for the below driver to complete, but just make
		//the call and go on, finishing cleanup
		IoCopyCurrentIrpStackLocationToNext( pIrp );
		Status = IoCallDriver( pStackDeviceObject, pIrp );
*/
		return STATUS_SUCCESS; //once was Status

	case IRP_MN_REMOVE_DEVICE:
		//The PnP Manager uses this IRP to direct drivers to remove a device. 
		//For a "polite" device removal, the PnP Manager sends an 
		//IRP_MN_QUERY_REMOVE_DEVICE prior to the remove IRP. In this case, 
		//the device is in the remove-pending state when the remove IRP arrives.
		//For a surprise-style device removal ( i.e. sudden cord yank ), 
		//the physical device has already been removed and the PnP Manager may not 
		//have sent IRP_MN_SURPRISE_REMOVAL. A device can be in any state
		//when it receives a remove IRP as a result of a surprise-style removal.

		//stop real driver
		AmUsbDisconnect( pDeviceObject->DeviceExtension );

		//decrement the FDO device extension's pending IO count for this IRP
		AmUsbDecIoCount( pDeviceObject );

		//Once DeviceRemoved is set no new IOCTL or read/write IRPs will be passed
		//down the stack to lower drivers; all will be quickly failed
		pDeviceExtension->DeviceRemoved = TRUE;

/* removed because this only generated bluescreens
		//We don't explicitly wait for the below driver to complete, but just make
		//the call and go on, finishing cleanup
		IoCopyCurrentIrpStackLocationToNext( pIrp );
		Status = IoCallDriver( pStackDeviceObject, pIrp );
*/
		//The final decrement to device extension PendingIoCount == 0
		//will set pDeviceExtension->RemoveEvent, enabling device removal.
		//If there is no pending IO at this point, the below decrement will be it.
		AmUsbDecIoCount( pDeviceObject );

		//wait for any IO request pending in our driver to
		//complete for finishing the remove
		KeWaitForSingleObject( 	&pDeviceExtension->evRemove,
					Suspended, KernelMode, FALSE, NULL );

		//delete the link and FDO we created
		AmUsbRemoveDevice( pDeviceObject );
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbPnp() Detaching from %08X\r\n", pDeviceExtension->pTopOfStackDeviceObject ) );
		IoDetachDevice( pDeviceExtension->pTopOfStackDeviceObject );
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbPnp() Deleting %08X\r\n", pDeviceObject ) );
		IoDeleteDevice( pDeviceObject );
		return STATUS_SUCCESS; //once was Status

	default:
		AmUsbDebug( DBGLVL_MAXIMUM, ( "AmUsbPnp() Minor PnP IOCTL not handled\r\n" ) );
	} //switch( pIrpStack->MinorFunction )

	//some error
	if( !NT_SUCCESS( Status ) )
	{
		//return failure without passing IRP down
		pIrp->IoStatus.Status = Status;
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );

		//decrement IO count for the FDO
		AmUsbDecIoCount( pDeviceObject );

		//debug message
		AmUsbDebug( DBGLVL_MAXIMUM, ( "AmUsbPnp() FAILED with status 0x%x\r\n", Status ) );

		return Status;
	}

	//copy IRP to next IRP
	IoCopyCurrentIrpStackLocationToNext( pIrp );

	//debug message
	AmUsbDebug( DBGLVL_MAXIMUM, ( "AmUsbPnp(): passing PnP IRP down, status = 0x%x\r\n", Status ) );

	//pass IRP down to lower PnP driver
	Status = IoCallDriver( pStackDeviceObject, pIrp );

	//decrement IO count for the FDO
	AmUsbDecIoCount( pDeviceObject );

	//debug message
	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbPnp() with status 0x%x\r\n", Status ) );

	return Status;
}

//initialize the functional device object
//pDriverObject: pointer to the driver object
//pPhysicalDeviceObject: pointer to the physical device object
//return value: STATUS_SUCCESS / STATUS_UNSECCESSFUL
NTSTATUS AmUsbPnpAddDevice( IN PDRIVER_OBJECT pDriverObject, IN PDEVICE_OBJECT pPhysicalDeviceObject )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PDEVICE_OBJECT pDeviceObject = NULL;
	PDEVICE_EXTENSION pDeviceExtension;
	USBD_VERSION_INFORMATION VersionInformation;
	ULONG i;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbPnpAddDevice()\r\n" ) );

	//create our funtional device object (FDO)
	Status = AmUsbCreateDeviceObject( pDriverObject, pPhysicalDeviceObject, &pDeviceObject );

	//success
	if( NT_SUCCESS( Status ) )
	{
		//get pointer to the device extension
		pDeviceExtension = pDeviceObject->DeviceExtension;
		//we are not initializing the device
		pDeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
		//we want buffered IO for reading and writing
		pDeviceObject->Flags |= DO_BUFFERED_IO;

		//Setting this flag causes the driver to not receive a IRP_MN_STOP_DEVICE
		//during suspend and also not get an IRP_MN_START_DEVICE during resume.  
		//This is neccesary because during the start device call, 
		//the GetDescriptors() call  will be failed by the USB stack.
		pDeviceObject->Flags |= DO_POWER_PAGABLE;

		//initialize our device extension
		pDeviceExtension->pPhysicalDeviceObject = pPhysicalDeviceObject; //remember the Physical device Object
		pDeviceExtension->pTopOfStackDeviceObject = IoAttachDeviceToDeviceStack( pDeviceObject, pPhysicalDeviceObject ); //attach to the PDO

		//no userspace app has opned this device
		pDeviceExtension->AppOpenCnt = 0;
		//no private device data yet
		pDeviceExtension->pDev = NULL;

		//Get a copy of the physical device's capabilities into a
		//DEVICE_CAPABILITIES struct in our device extension;
		//We are most interested in learning which system power states
		//are to be mapped to which device power states for handling
		//IRP_MJ_SET_POWER IRPs.
		AmUsbQueryCapabilities( pPhysicalDeviceObject, &pDeviceExtension->DeviceCapabilities );

		//We want to determine what level to auto-powerdown to; This is the lowest
		//sleeping level that is LESS than D3; 
		//If all are set to D3, auto powerdown/powerup will be disabled.
		pDeviceExtension->PowerDownLevel = PowerDeviceUnspecified; //init to disabled
		for( i = PowerSystemSleeping1; i <= PowerSystemSleeping3; i++ )
			if( pDeviceExtension->DeviceCapabilities.DeviceState[i] < PowerDeviceD3 )
				pDeviceExtension->PowerDownLevel = pDeviceExtension->DeviceCapabilities.DeviceState[i];

		//display the device  caps
		AmUsbDebug( DBGLVL_MEDIUM, ( " >>>>>> DeviceCaps\r\n" ) );
		AmUsbDebug( DBGLVL_MEDIUM, ( " SystemWake = %s\r\n",
				AmUsbStringForSysState( pDeviceExtension->DeviceCapabilities.SystemWake ) ) );
		AmUsbDebug( DBGLVL_MEDIUM, ( " DeviceWake = %s\r\n",
				AmUsbStringForDevState( pDeviceExtension->DeviceCapabilities.DeviceWake) ) );
		for( i = PowerSystemUnspecified; i < PowerSystemMaximum; i++ )
			AmUsbDebug( DBGLVL_MEDIUM, ( " Device State Map: sysstate %s = devstate %s\r\n",
					AmUsbStringForSysState( i ),
					AmUsbStringForDevState( pDeviceExtension->DeviceCapabilities.DeviceState[i] ) ) );
		AmUsbDebug( DBGLVL_MEDIUM, ( " <<<<<<<<DeviceCaps\r\n" ) );

		//We keep a pending IO count ( extension->PendingIoCount )  in the device extension.
		//The first increment of this count is done on adding the device.
		//Subsequently, the count is incremented for each new IRP received and
		//decremented when each IRP is completed or passed on.
		//Transition to 'one' therefore indicates no IO is pending and signals
		//pDeviceExtension->NoPendingIoEvent. This is needed for processing
		//IRP_MN_QUERY_REMOVE_DEVICE
		//Transition to 'zero' signals an event ( pDeviceExtension->RemoveEvent )
		//to enable device removal. This is used in processing for IRP_MN_REMOVE_DEVICE
		AmUsbIncIoCount( pDeviceObject );
	} //if( NT_SUCCESS( Status ) )

	//get version of USB bus driver
	USBD_GetUSBDIVersion( &VersionInformation );

	//success
	if( NT_SUCCESS( Status ) )  
	{
		NTSTATUS ActStat;
		//try to power down device until IO actually requested
		ActStat = AmUsbSelfSuspendOrActivate( pDeviceObject, TRUE );
	}

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbPnpAddDevice() with status 0x%x\r\n", Status ) );

	return Status;
}
